@@ -12,11 +12,7 @@ from math import *
 from math import _greater_than, _permutation_by_sort, _permute, _approx_f_n, _gamma_KL, _dirichlet_KL, _beta_KL, _safe_x_log_x
 from utils import *
 from summarise import *
-
-
-
-
-
+import types
 
 
 class ApproxCounts(object):
@@ -94,7 +90,7 @@ class ApproxCounts(object):
 
     def __str__(self):
         return 'ApproxCounts object'
-    
+
     def permute(self, permutation):
         "Permute the counts based on a permutation of the topics."
         self.E_n_dk = _permute(self.E_n_dk, permutation, axis=1)
@@ -144,11 +140,11 @@ class HDPM(object):
     Implementation of hierarchical Dirichlet processes as detailed in
     http://www.gatsby.ucl.ac.uk/~ywteh/research/inference/nips2007b.pdf
     """
-    
+
     INIT_METHOD_SIMPLE = 1
     INIT_METHOD_SAMPLE = 2
-    
-    
+
+
     def _memoize_calculations(self):
         """
         Memoize those calculations we might otherwise make over and over.
@@ -161,20 +157,33 @@ class HDPM(object):
         self._calculate_counts      = memoize(self._calculate_counts)
 
 
+    def __getstate__(self):
+        state = {}
+        for attr, value in self.__dict__.iteritems():
+            if not isinstance(value, Memoize):
+                state[attr] = value
+        return state
+
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        self._memoize_calculations()
+
+
     def __init__(
-      self,
-      documents,
-      W,
-      K,
-      a_alpha = 4.,
-      b_alpha = 4.,
-      a_beta = 4.,
-      b_beta = 4.,
-      a_gamma = 7.5,
-      b_gamma = 7.5,
-      a_tau = None,
-      init_method = INIT_METHOD_SAMPLE,
-      init_method_args = None
+        self,
+        documents,
+        W,
+        K,
+        a_alpha = 4.,
+        b_alpha = 4.,
+        a_beta = 4.,
+        b_beta = 4.,
+        a_gamma = 7.5,
+        b_gamma = 7.5,
+        a_tau = None,
+        init_method = INIT_METHOD_SAMPLE,
+        init_method_args = None
     ):
         """
         Initialise the hierarchical Dirichlet process mixture.
@@ -234,7 +243,7 @@ class HDPM(object):
 
         self.D = len(self.documents)
         "Number of documents"
-        
+
         self.n_d = numpy.array([len(document) for document in self.documents])
         "Number of words in each document"
 
@@ -254,7 +263,7 @@ class HDPM(object):
 
     def initialise_hdpm_variational_simply(self, q_z_dirichlet_param = 1.):
         "Old straightforward way to initialise the variational distributions."
-        
+
         self.q_alpha = GammaDist(self.a_alpha, self.b_alpha)
         "Variational distribution over alpha"
 
@@ -277,13 +286,13 @@ class HDPM(object):
         ]
         "Variational distribution over z_id, indexed by d."
         self._normalise_z()
-        
+
 
     def initialise_hdpm_variational_by_sampling(self):
         """
         Initialise the variational distributions of a HDPM by sampling the generative model (ignoring the data).
         """
-        
+
         self.q_gamma = GammaDist(self.a_gamma, self.b_gamma)
         "Variational distribution over gamma"
         gamma = self.q_gamma.sample()
@@ -297,12 +306,14 @@ class HDPM(object):
         pi[-1] = 1. - pi[:-1].sum()
         if pi[-1] < 0.: # adjust for numerical errors
             pi[-1] = 0.
-        
+
         self.q_alpha = GammaDist(self.a_alpha, self.b_alpha)
         "Variational distribution over alpha"
         alpha = self.q_alpha.sample()
 
         alpha_pi = alpha * pi
+        for document in self.documents:
+            numpy.random.dirichlet(alpha_pi, size=len(document))
         self.q_z = [
             numpy.random.dirichlet(alpha_pi, size=len(document))
             for document in self.documents
@@ -317,8 +328,8 @@ class HDPM(object):
         "Variational distribution over tau"
 
         return gamma, pi_bar, pi, alpha
-    
-    
+
+
     def _calculate_E_s_dk(self):
         G_alpha_pi = self.q_alpha.G * self.G_pi
         term = G_alpha_pi + self.counts.E_plus_n_dk
