@@ -15,21 +15,24 @@ from cookbook import pylab_utils
 from cookbook.pylab_utils import pylab_ioff
 
 
-def heatmap_categories(categories, category_names=None, distribution_names=None):
+def heatmap_categories(
+        categories, 
+        category_names=None, 
+        distribution_names=None):
     "Plot the categories as a heat map."
-    
+
     import pylab as P
     assert 2 == len(categories.shape)
 
     num_dists = categories.shape[0]
     assert None == distribution_names or len(distribution_names) == num_dists
-    
+
     num_categories = categories.shape[1]
     assert None == category_names or len(category_names) == num_categories
 
     x_ind = numpy.arange(num_categories)
     y_ind = numpy.arange(num_dists)
-    
+
     plt = P.pcolor(categories)
     P.colorbar()
     axes = P.gca()
@@ -52,8 +55,8 @@ def heatmap_categories(categories, category_names=None, distribution_names=None)
         t.tick1On = False
         t.tick2On = False
 
-    
-    
+
+
 
 
 def plot_categories(categories, category_names=None, distribution_names=None, numrows=None, numcols=None):
@@ -61,7 +64,7 @@ def plot_categories(categories, category_names=None, distribution_names=None, nu
 
     import pylab as P
     assert 2 == len(categories.shape)
-    
+
     # determine number of rows and columns
     num_dists = categories.shape[0]
     num_categories = categories.shape[1]
@@ -263,7 +266,7 @@ class DpmSummariser(object):
         @arg word_tag: A string to replace the word "word" with in the output.
         @arg occurence_tag: A string to replace the word "occurence" with in the output.
         """
-        
+
         if None == document_ids:
             document_ids = [str(i) for i in xrange(dpm.D)]
         if None == word_ids:
@@ -499,10 +502,13 @@ class DpmSummariser(object):
         logging.info('Creating topic size histogram (log scale)')
         import pylab as P
         P.figure()
-        P.bar(xrange(self.statistics.num_topics_used), self.dpm.counts.E_n_k[:self.statistics.num_topics_used], log=True)
+        P.bar(
+             xrange(self.statistics.num_topics_used),
+             numpy.log10(
+                 self.dpm.counts.E_n_k[:self.statistics.num_topics_used]))
         P.title('%s sizes (log scale)' % self.topic_tag)
         P.xlabel('programs')
-        P.ylabel('sizes')
+        P.ylabel('log10 sizes')
         self.save_fig('%s-sizes-log-scale' % self.topic_tag)
         P.close()
 
@@ -561,6 +567,7 @@ class DpmSummariser(object):
         """
         Create a figure of all histograms
         """
+        logging.info('Creating word/topic/document multi-histogram')
         import pylab as P
         fig = P.figure()
         P.subplot(2,2,1)
@@ -595,6 +602,7 @@ class DpmSummariser(object):
         self.save_fig('associations')
         P.close()
 
+
     @pylab_ioff
     def _plot_dist(self, dist, title, tag):
         import pylab as P
@@ -603,20 +611,111 @@ class DpmSummariser(object):
         self.save_fig(tag)
         P.close()
 
+
     def _dist_plots(self):
         self._plot_dist(self.dpm.q_alpha, 'Alpha', 'alpha')
         self._plot_dist(self.dpm.q_beta, 'Beta', 'beta')
         self._plot_dist(self.dpm.q_gamma, 'Gamma', 'gamma')
 
+
     def _log_hyper_info(self, p, name):
-        logging.info('%8s: E=%4g; G=%4g; params: %s', name, p.E, p.G, str(p.params()))
+        logging.info(
+            '%8s: E=%4g; G=%4g; params: %s', 
+            name, p.E, p.G, str(p.params()))
+
+
+    def _plot_hyper_dist(self, p, prior_params, name):
+        logging.info(
+            '%8s: E=%4g; G=%4g; params: %s', 
+            name, p.E, p.G, str(p.params()))
+        prior = p.__class__(*prior_params)
+        xmax = max(2 * p.E, 2 * prior.E)
+        import pylab as P
+        p.plot(num_points=150, xmax=xmax, label='posterior')
+        prior.plot(num_points=150, xmax=xmax, label='prior')
+        P.legend()
+
+
+    @pylab_ioff
+    def plot_hyper_parameters(self):
+        """Plot hyper-parameters."""
+        import pylab as P
+        fig = P.figure(figsize=(12,9))
+        P.subplot(2,2,1)
+        self._plot_hyper_dist(
+            self.dpm.q_alpha,
+            (self.dpm.a_alpha, self.dpm.b_alpha),
+            'alpha')
+        P.subplot(2,2,2)
+        self._plot_hyper_dist(
+            self.dpm.q_beta,
+            (self.dpm.a_beta, self.dpm.b_beta),
+            'beta')
+        P.subplot(2,2,3)
+        self._plot_hyper_dist(
+            self.dpm.q_gamma,
+            (self.dpm.a_gamma, self.dpm.b_gamma),
+            'gamma')
+        P.title('Hyperparameters')
+        self.save_fig('hyperparameters')
+        P.close()
+
 
     def log_hyper_parameter_info(self):
-        'Log some information about the distributions over the hyper-parameters.'
+        """Log some information about the distributions over
+        the hyper-parameters."""
         self._log_hyper_info(self.dpm.q_alpha, 'Alpha')
         self._log_hyper_info(self.dpm.q_beta, 'Beta')
         self._log_hyper_info(self.dpm.q_gamma, 'Gamma')
 
+
+    def build_graph(self):
+        logging.info('Building graph')
+        import graph_tool.all as GT
+        graph = GT.Graph(directed=False)
+        vertices = dict()
+        names = graph.new_vertex_property('string')
+        shapes = graph.new_vertex_property('string')
+        colours = graph.new_vertex_property('string')
+        def get_vertex(name, shape, colour):
+            if name in vertices:
+                return vertices[name]
+            else:
+                vertex = graph.add_vertex()
+                names[vertex] = name
+                shapes[vertex] = shape
+                colours[vertex] = colour
+                vertices[name] = vertex
+                return vertex
+
+        def get_word_vertex(name):
+            return get_vertex(name, 'circle', 'blue')
+
+        def get_topic_vertex(name):
+            return get_vertex(name, 'square', 'red')
+
+        def get_document_vertex(name):
+            return get_vertex(name, 'triangle', 'green')
+
+        for k in xrange(self.statistics.num_topics_used):
+            name = 'TP: %d' % k
+            topic_vertex = get_topic_vertex(name)
+            for w in self.statistics.words_for_topic(k):
+                word_vertex = get_word_vertex('Word: %d' % w)
+                graph.add_edge(word_vertex, topic_vertex)
+            for d in self.statistics.documents_for_topic(k):
+                doc_vertex = get_document_vertex('Doc: %d' % d)
+                graph.add_edge(topic_vertex, doc_vertex)
+
+        pos = GT.sfdp_layout(graph)
+        GT.graph_draw(
+            graph,
+            pos=pos,
+            vertex_shape=shapes,
+            vertex_fill_color=colours)
+
+        return graph, vertices, shapes, colours
+
     def log_topic_info(self, k):
         'Log some general information about topic k.'
         topic_words = self.statistics.words_for_topic(k)
@@ -649,14 +748,107 @@ class DpmSummariser(object):
         return fig
 
 
+    def write_posterior(self):
+        """
+        Writes the expected phi theta to disk.
+        """
+        phi = self.dpm.exp_phi()
+        assert phi.shape[0] == self.dpm.K
+        with open('%s-phi.txt' % self.prefix, 'w') as out:
+            print >>out, ','.join(self.word_ids)
+            for k in xrange(self.dpm.K):
+                print >>out, ','.join(map(str, phi[k]))
+
+        theta = self.dpm.exp_theta()
+        assert theta.shape[1] == self.dpm.K
+        with open('%s-theta.txt' % self.prefix, 'w') as out:
+            print >>out, ','.join(self.document_ids)
+            for k in xrange(self.dpm.K):
+                print >>out, ','.join(map(str, theta[:,k]))
+
+
+    def write_posterior_enrichment(self):
+        """
+        Writes the ratios of the expected phi/Phi and theta/Theta to disk.
+        """
+        phi = self.dpm.exp_phi()
+        Phi = self.dpm.exp_Phi()
+        phi_ratio = phi/Phi
+        assert phi_ratio.shape[0] == self.dpm.K
+        with open('%s-phi-ratio.txt' % self.prefix, 'w') as out:
+            print >>out, ','.join(self.word_ids)
+            for k in xrange(self.dpm.K):
+                print >>out, ','.join(map(str, phi_ratio[k]))
+
+        theta = self.dpm.exp_theta()
+        Theta = self.dpm.exp_Theta()
+        theta_ratio = theta/Theta
+        assert theta_ratio.shape[1] == self.dpm.K
+        with open('%s-theta-ratio.txt' % self.prefix, 'w') as out:
+            print >>out, ','.join(self.document_ids)
+            for k in xrange(self.dpm.K):
+                print >>out, ','.join(map(str, theta_ratio[:,k]))
+
+
 
 class DpmInferenceHistory(object):
     "Keeps track of statistics during inference."
 
+    def __getstate__(self):
+        """Pickling helper function."""
+        return (self.dpm, self.summariser, self.history)
+
+
+    def __setstate__(self, state):
+        """Pickling helper function."""
+        self.dpm, self.summariser, self.history = state
+        self._init_stat_fns()
+
+
+    def _init_stat_fns(self):
+        self.stats = {
+            'hyperparameters' : {
+                'alpha' : {
+                     'E(alpha)' : lambda: self.dpm.q_alpha.E,
+                     'G(alpha)' : lambda: self.dpm.q_alpha.G,
+                },
+                'beta' : {
+                     'E(beta)'  : lambda: self.dpm.q_beta.E,
+                     'G(beta)'  : lambda: self.dpm.q_beta.G,
+                },
+                'gamma' : {
+                     'E(gamma)' : lambda: self.dpm.q_gamma.E,
+                     'G(gamma)' : lambda: self.dpm.q_gamma.G,
+                },
+                'num_topics_used' : {
+                     '\\# topics used' : lambda: self.summariser.statistics.num_topics_used,
+                },
+            },
+            'log-likelihood' : {
+                'log-likelihood' : {
+                    'negative log likelihood' : lambda: -self.dpm.log_likelihood(),
+                },
+            },
+            'topic-sizes' : {
+                'topic-allocations' : {
+                   'E(topics/word)'     : lambda: numpy.mean(self.summariser.statistics.num_topics_per_word()),
+                   'E(topics/document)' : lambda: numpy.mean(self.summariser.statistics.num_topics_per_document()),
+                },
+                'topic-sizes' : {
+                   'E(words/topic)'     : lambda: numpy.mean(self.summariser.statistics.num_words_per_topic()),
+                   'E(documents/topic)' : lambda: numpy.mean(self.summariser.statistics.num_documents_per_topic()),
+                },
+            },
+        }
+        """A dictionary of dictionaries of dictionaries.
+        Top level keys are figure names.
+        Second level keys are plot names.
+        Third level keys are stats in that plot.
+        """
+
+
     def __init__(self, dpm, summariser):
-        
-        self.plot = pylab_ioff(self.plot)
-        
+
         self.dpm = dpm
         "The DPM."
 
@@ -666,45 +858,34 @@ class DpmInferenceHistory(object):
         self.history = cookbook.DictOfLists()
         "Holds the history of each statistic by name."
 
-        self.stats = {
-          'log-likelihood' : {
-            'negative log likelihood' : lambda: -self.dpm.log_likelihood(),
-          },
-          'hyper-parameters' : {
-             'E(alpha)' : lambda: self.dpm.q_alpha.E,
-             'G(alpha)' : lambda: self.dpm.q_alpha.G,
-             'E(beta)'  : lambda: self.dpm.q_beta.E,
-             'G(beta)'  : lambda: self.dpm.q_beta.G,
-             'E(gamma)' : lambda: self.dpm.q_gamma.E,
-             'G(gamma)' : lambda: self.dpm.q_gamma.G,
-          },
-          'topic-sizes' : {
-             'E(words/topic)'     : lambda: numpy.mean(self.summariser.statistics.num_words_per_topic()),
-             'E(topics/word)'     : lambda: numpy.mean(self.summariser.statistics.num_topics_per_word()),
-             'E(topics/document)' : lambda: numpy.mean(self.summariser.statistics.num_topics_per_document()),
-             'E(documents/topic)' : lambda: numpy.mean(self.summariser.statistics.num_documents_per_topic()),
-          },
-          'num_topics_used' : {
-             '\\# topics used' : lambda: self.summariser.statistics.num_topics_used,
-          },
-        }
-        "A dictionary of dictionaries. Top level keys are plot names, second level keys are stats in that plot."
+        self._init_stat_fns()
+
 
     def iteration(self):
-        for plot, stats in self.stats.iteritems():
-            for name, fn in stats.iteritems():
-                self.history[name].append(fn())
+        for figname, figurestats in self.stats.iteritems():
+            for subplotname, subplotstats in figurestats.iteritems():
+                for name, fn in subplotstats.iteritems():
+                    self.history[name].append(fn())
         return -self.history['negative log likelihood'][-1]
 
+
     @pylab_ioff
     def plot(self):
         import pylab as P
-        for plot, stats in self.stats.iteritems():
-            P.figure()
-            for name, fn in stats.iteritems():
-                P.plot(self.history[name], label=name)
-            P.legend()
-            self.summariser.save_fig(tag='history-%s' % plot)
+        for figname, figurestats in self.stats.iteritems():
+            P.figure(figsize=(12,9))
+            numsubplots = len(figurestats)
+            nrows = int(numpy.sqrt(numsubplots))
+            ncols = int(numsubplots / nrows)
+            while nrows * ncols < numsubplots:
+                ncols += 1
+            for p, (subplotname, subplotstats) in enumerate(figurestats.iteritems()):
+                P.subplot(nrows, ncols, p+1)
+                for name, fn in subplotstats.iteritems():
+                    P.plot(self.history[name], label=name)
+                P.legend()
+                P.xlabel('Iteration')
+            self.summariser.save_fig(tag='history-%s' % figname)
             P.close()
 
 
