@@ -1,5 +1,5 @@
 #
-# Copyright John Reid 2007, 2008, 2009
+# Copyright John Reid 2007, 2008, 2009. 2013
 #
 
 """
@@ -8,12 +8,16 @@ Code to implement ROC point/curve calculation and plotting.
 
 
 import math, numpy as N
-from itertools import chain
+from itertools import chain, groupby
+import logging
+import warnings
+logging.basicConfig(level=logging.INFO)
 
 
-class RocCalculator( object ):
+class RocCalculator(object):
     """
-    Calculates specificities and sensitivities
+    Calculates specificities and sensitivities from counts of true and false
+    positives and negatives.
 
     Source: wikipedia - Fawcett (2004)
     """
@@ -32,6 +36,18 @@ class RocCalculator( object ):
         "Number of false negatives."
 
 
+    @property
+    def total_positive(self):
+        "The total number of positive test cases."
+        return self.tp + self.fn
+
+
+    @property
+    def total_negative(self):
+        "The total number of negative test cases."
+        return self.tn + self.fp
+
+
     def __cmp__(self, other):
         "Comparison."
         diff = self.sensitivity() - other.sensitivity()
@@ -63,6 +79,7 @@ class RocCalculator( object ):
             if truth: self.fn += 1
             else: self.tn += 1
 
+
     def __add__(self, rhs):
         "Add this RocCalculator to another and return the result."
         result = RocCalculator()
@@ -72,6 +89,7 @@ class RocCalculator( object ):
         result.fn = self.fn + rhs.fn
         return result
 
+
     def normalise(self, rhs):
         "Normalise this RocCalculator so that tp+tn+fp+fn=1."
         sum = float(self.tp+self.tn+self.fp+self.fn)
@@ -80,6 +98,7 @@ class RocCalculator( object ):
         self.fp /= sum
         self.fn /= sum
 
+
     @staticmethod
     def always_predict_true():
         "A RocCalculator for a predictor that always predicts True"
@@ -88,6 +107,7 @@ class RocCalculator( object ):
         result.tn = result.fn = 0
         return result
 
+
     @staticmethod
     def always_predict_false():
         "A RocCalculator for a predictor that always predicts False"
@@ -96,6 +116,7 @@ class RocCalculator( object ):
         result.tn = result.fn = 1
         return result
 
+
     def sensitivity(self):
         "TP/(TP+FN)"
         denominator = self.tp + self.fn
@@ -103,7 +124,8 @@ class RocCalculator( object ):
             return float(self.tp)/denominator
         else:
             return 1.0
-    hit_rate = recall = sensitivity
+    true_positive_rate = tpr = hit_rate = recall = sensitivity
+
 
     def specificity(self):
         "TN/(TN+FP)"
@@ -113,9 +135,12 @@ class RocCalculator( object ):
         else:
             return 1.0
 
+
     def false_positive_rate(self):
         "FP/(TN+FP)"
         return 1.0 - self.specificity()
+    fpr = false_positive_rate
+
 
     def positive_predictive_value(self):
         "TP/(TP+FP)"
@@ -126,6 +151,7 @@ class RocCalculator( object ):
             return 1.0
     precision = positive_predictive_value
 
+
     def negative_predictive_value(self):
         "TN/(TN+FN)"
         denominator = self.tn + self.fn
@@ -142,6 +168,7 @@ class RocCalculator( object ):
         else:
             return 1.0
 
+
     def accuracy(self):
         "(TP+TN)/(TP+TN+FP+FN)"
         denominator = self.tp + self.tn + self.fn + self.fp
@@ -149,11 +176,13 @@ class RocCalculator( object ):
             return float(self.tp+self.tn)/denominator
         else:
             return 1.0
-    
+
+
     def average_performance(self):
         "(sensitivity()+positive_predictive_value())/2"
         return (self.sensitivity() + self.positive_predictive_value()) / 2.
 
+
     def correlation_coefficient(self):
         "(TP.TN-FN.FP)/sqrt((TP+FN)(TN+FP)(TP+FP)(TN+FN)) see: Burset & Guigo"
         denominator = math.sqrt((self.tp+self.fn)*(self.tn+self.fn)*(self.tp+self.fp)*(self.tn+self.fp))
@@ -166,6 +195,7 @@ class RocCalculator( object ):
             else:
                 return 1.0
 
+
     def __str__(self):
         return '''TP: %d; FP: %d; TN: %d; FN: %d
 sensitivity:               %.3f    TP/(TP+FN)
@@ -181,13 +211,13 @@ self.tp, self.fp, self.tn, self.fn,
                 self.correlation_coefficient(),
         )
 
+
 def update_roc(roc, truth_prediction_iterable):
     "for each (truth,prediction) in iterable, update the ROC calculator"
     for truth, prediction in truth_prediction_iterable:
         roc(truth, prediction)
 
 
-
 def get_new_roc_parameter(rocs, for_specificity=True):
     """
     Takes a sequence of (parameter, roc) tuples and returns a new parameter that should be tested
@@ -207,6 +237,24 @@ def get_new_roc_parameter(rocs, for_specificity=True):
     return max(diffs)[1]
 
 
+def generate_roc_points(rocs, sort_negative_first=True):
+    """
+    Generate ROC points but sort negatives before positives at same
+    threshold if asked to.
+    This gives a step-function like ROC curve rather than a smoothed curve.
+    """
+    warnings.warn('DEPRECATED: use infpy.roc.all_rocs_from_thresholds()', DeprecationWarning)
+    last = RocCalculator.always_predict_false()
+    yield last
+    for roc in chain(rocs, (RocCalculator.always_predict_true(),)):
+        if sort_negative_first:
+            # add another ROC point with TPR same as last ROC point but FPR
+            # same as this one
+            yield RocCalculator(last.tp, roc.fp, roc.tn, last.fn)
+        yield roc
+        last = roc
+
+
 def plot_roc_points(rocs, **plot_kwds):
     """
     Plots TPR versus FPR for the ROCs in rocs. Adds points at (0,0) and (1,1).
@@ -215,12 +263,28 @@ def plot_roc_points(rocs, **plot_kwds):
     :param plot_kwds: All extra keyword arguments are passed to the pylab.plot call.
     :returns: The result of pylab.plot call.
     """
+    warnings.warn('DEPRECATED: use infpy.roc.plot_rocpoints()', DeprecationWarning)
     from pylab import plot
-    return plot(
-     [s for s in chain((0.,), (1. - roc.specificity() for roc in rocs), (1.,))],
-     [s for s in chain((0.,), (roc.sensitivity() for roc in rocs), (1.,))],
-     **plot_kwds
-    )
+    extended_rocs = list(generate_roc_points(rocs))
+    tprs = map(RocCalculator.tpr, extended_rocs)
+    fprs = map(RocCalculator.fpr, extended_rocs)
+    return plot(fprs, tprs, **plot_kwds)
+
+
+def plot_rocpoints(rocpoints, fillargs=None, **plot_kwds):
+    """
+    Plots TPR versus FPR for the ROCs in rocpoints.
+
+    :param rocpoints: A sequence of ROCs.
+    :param plot_kwds: All extra keyword arguments are passed to the pylab.plot call.
+    :returns: The result of pylab.plot call.
+    """
+    from pylab import plot, fill_between
+    tprs = map(RocCalculator.tpr, rocpoints)
+    fprs = map(RocCalculator.fpr, rocpoints)
+    if fillargs is not None:
+        fill_between(fprs, tprs, **fillargs)
+    return plot(fprs, tprs, **plot_kwds)
 
 
 def plot_precision_versus_recall(rocs, **plot_kwds):
@@ -235,9 +299,9 @@ def plot_precision_versus_recall(rocs, **plot_kwds):
     points = [(roc.recall(), roc.precision()) for roc in rocs]
     #points.sort()
     return plot(
-            [recall for recall, precision in points],
-            [precision for recall, precision in points],
-            **plot_kwds
+        [recall for recall, precision in points],
+        [precision for recall, precision in points],
+        **plot_kwds
     )
 
 
@@ -279,6 +343,22 @@ def plot_precision_recall(roc_thresholds, recall_plot_kwds={}, precision_plot_kw
     return recall_result, precision_result
 
 
+def auc(rocpoints):
+    """
+    Calculate the area under the ROC points.
+    """
+    sum = 0.
+    last = None
+    for point in rocpoints:
+        try:
+            sum += ((point.tpr() + last.tpr())
+                / 2 * abs(last.fpr() - point.fpr()))
+        except AttributeError:
+            pass
+        last = point
+    return sum
+
+
 def area_under_curve(rocs, include_0_0=True, include_1_1=True):
     """
     :param rocs: The ROC points.
@@ -287,6 +367,7 @@ def area_under_curve(rocs, include_0_0=True, include_1_1=True):
 
     :returns: The area under the ROC curve given by the ROC points.
     """
+    warnings.warn('DEPRECATED: use infpy.roc.auc()', DeprecationWarning)
     x_axis = []
     y_axis = []
     if include_0_0:
@@ -314,10 +395,10 @@ def plot_random_classifier(**kwargs):
     if 'linestyle' not in kwargs:
         kwargs['linestyle'] = ':'
     plot(
-    [0,1],
-    [0,1],
-    **kwargs
-)
+        [0,1],
+        [0,1],
+        **kwargs
+    )
 
 
 def label_plot():
@@ -378,12 +459,13 @@ def rocs_from_thresholds(positive_thresholds, negative_thresholds, num_points=32
 
     :returns: A list of ROC points.
     """
+    warnings.warn('DEPRECATED: use infpy.roc.all_rocs_from_thresholds()', DeprecationWarning)
     min_threshold = min(positive_thresholds[0], negative_thresholds[0])
     max_threshold = max(positive_thresholds[-1], negative_thresholds[-1])
     rocs = map(
-                make_roc_from_threshold_fn(positive_thresholds, negative_thresholds),
-                N.linspace(min_threshold, max_threshold, num_points)[::-1]
-        )
+        make_roc_from_threshold_fn(positive_thresholds, negative_thresholds),
+        N.linspace(min_threshold, max_threshold, num_points)[::-1]
+    )
     return rocs
 
 
@@ -423,12 +505,13 @@ def create_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_po
 
     :returns: A list of tuples (ROC point, threshold).
     """
+    warnings.warn('DEPRECATED: use infpy.roc.all_rocs_from_thresholds()', DeprecationWarning)
     return pick_roc_thresholds(
-                make_roc_from_threshold_fn(positive_thresholds, negative_thresholds),
-            min_threshold=min(positive_thresholds[0], negative_thresholds[0]),
-            max_threshold=max(positive_thresholds[-1], negative_thresholds[-1]),
-            num_points=num_points
-        )
+        make_roc_from_threshold_fn(positive_thresholds, negative_thresholds),
+        min_threshold=min(positive_thresholds[0], negative_thresholds[0]),
+        max_threshold=max(positive_thresholds[-1], negative_thresholds[-1]),
+        num_points=num_points
+    )
 
 
 def picked_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_points=32):
@@ -438,9 +521,75 @@ def picked_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_po
 
     :returns: A list of ROC points.
     """
+    warnings.warn('DEPRECATED: use infpy.roc.all_rocs_from_thresholds()', DeprecationWarning)
     return [roc for roc, t in create_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_points=num_points)]
 
 
+def all_rocs_from_thresholds(
+    positive_thresholds,
+    negative_thresholds,
+    negative_first=True
+):
+    """
+    Takes 2 sorted lists (smallest to largest):
+    one list is of the thresholds required to classify the positive examples
+    as positive and the other list is of the thresholds required to classify
+    the negative examples as positive.
+
+    :returns: Yields all the ROC points. Note that they are returned in the
+        opposite order to some of the other methods in this module.
+    """
+    import heapq
+
+    #
+    # How many do we have in total?
+    #
+    total_positive = len(positive_thresholds)
+    total_negative = len(negative_thresholds)
+    if not total_positive:
+        raise ValueError('Need to have at least one positive prediction')
+    if not total_negative:
+        raise ValueError('Need to have at least one negative prediction')
+
+    #
+    # At the lowest threshold, everything is a positive prediction
+    #
+    tp = total_positive
+    fp = total_negative
+    yield RocCalculator(tp, fp, 0, 0)
+
+    #
+    # Iterate through the merged thresholds.
+    #
+    if not negative_first:
+        keyfn = lambda x: x[0] # just group by threshold not positive/negative
+    else:
+        keyfn = None
+    for key, group in groupby(
+        heapq.merge(
+            ((t, 1) for t in negative_thresholds),
+            ((t, 0) for t in positive_thresholds)),
+        key=keyfn
+    ):
+        #
+        # Update our number of true or false positive predictions
+        #
+        for threshold, isnegative in group:
+            if isnegative:
+                fp -= 1
+            else:
+                tp -= 1
+
+        #
+        # Yield the ROC point for this group
+        #
+        yield RocCalculator(tp, fp,
+                            total_negative-fp, total_positive-tp)
+
+    # everything should be classified as false after seeing all the thresholds
+    assert tp == 0
+    assert fp == 0
+
 
 def resize_negative_examples(positive_thresholds, negative_thresholds, num_negative=50):
     """
@@ -461,6 +610,7 @@ def auc50_wrong(positive_thresholds, negative_thresholds, num_negative=50, num_p
     """
     Calculate the AUC50 as in Gribskov & Robinson 'Use of ROC analysis to evaluate sequence pattern matching'
     """
+    warnings.warn('DEPRECATED: use infpy.roc.auc50_from_rocpoints()', DeprecationWarning)
     if num_negative > len(negative_thresholds):
         raise RuntimeError('Not enough negative examples (%d). Requested %d' % (len(negative_thresholds), num_negative))
     threshold = negative_thresholds[-num_negative]
@@ -474,10 +624,16 @@ def auc50_wrong(positive_thresholds, negative_thresholds, num_negative=50, num_p
     return auc50, roc_thresholds
 
 
-def auc50(positive_thresholds, negative_thresholds, num_negative=50, num_points=32):
+def auc50(
+    positive_thresholds,
+    negative_thresholds,
+    num_negative=50,
+    num_points=32):
     """
-    Calculate the AUC50 as in Gribskov & Robinson 'Use of ROC analysis to evaluate sequence pattern matching'
+    Calculate the AUC50 as in Gribskov & Robinson:
+    'Use of ROC analysis to evaluate sequence pattern matching'
     """
+    warnings.warn('DEPRECATED: use infpy.roc.auc50_from_rocpoints()', DeprecationWarning)
     if num_negative > len(negative_thresholds):
         raise RuntimeError('Not enough negative examples (%d). Requested %d' % (len(negative_thresholds), num_negative))
     roc_thresholds = pick_roc_thresholds(
@@ -490,58 +646,180 @@ def auc50(positive_thresholds, negative_thresholds, num_negative=50, num_points=
     return auc50, roc_thresholds
 
 
+def plot_rocpoint(rocpoint, **plotargs):
+    """Plot a single rocpoint. Typically used to indicate where the last point for the AUC50
+    calculation is."""
+    from pylab import plot
+    plot(
+        [rocpoint.fpr()],
+        [rocpoint.tpr()],
+        color='black',
+        marker='x',
+        markersize=6,
+        **plotargs
+    )
+
+
+def bisect_rocs(rocpoints, predicate, start=0, end=None):
+    """Return the index into rocpoints for first rocpoint with
+    predicate(rocpoint) is True and start <= index < end. Assumes
+    rocpoints are sorted w.r.t. predicate.
+    """
+    if end is None:
+        end = len(rocpoints)
+    if end <= start:
+        raise ValueError('Start (%d) must be less than end (%d).' % (start, end))
+    new_index = (start + end) / 2
+    if start == new_index:
+        return end  # We narrowed down the range completely
+    else:
+        if predicate(rocpoints[new_index]):
+            # Look in lower half
+            return bisect_rocs(rocpoints, predicate, start, new_index)
+        else:
+            # Look in upper half
+            return bisect_rocs(rocpoints, predicate, new_index, end)
+
+
+def restrict_false_positives(rocpoints, max_fp=50):
+    """
+    Yield the ROC points while the number of true negatives is less than
+    max_tn.
+    """
+    last = None
+    for point in rocpoints:
+        # check the rocpoints are in the correct sorted order
+        try:
+            if last.fp > point.fp:
+                raise ValueError(
+                    'ROC points should be ordered by increasing '
+                    'false positives')
+        except AttributeError:
+            pass
+        if point.fp <= max_fp:
+            yield point
+            if point.fp == max_fp:
+                break
+        else:
+            yield RocCalculator(
+                point.tp, max_fp, point.tn - max_fp + point.fp, point.fn)
+            break
+        last = point
+
+
+def auc50_from_rocpoints(
+    rocpoints,
+    max_fp=50):
+    """
+    Calculate the AUC50 as in Gribskov & Robinson:
+    'Use of ROC analysis to evaluate sequence pattern matching'
+    """
+    return auc(restrict_false_positives(rocpoints, max_fp))
+
+
 if '__main__' == __name__:
     import numpy.random as R, pylab as P
 
-    num_examples = 1000
-    R.seed(2)
-    positive_thresholds = R.normal(size=num_examples, loc=.4, scale=2.)
-    positive_thresholds.sort()
-    negative_thresholds = R.normal(size=num_examples, loc=-.4, scale=1.)
-    negative_thresholds.sort()
+    def check_points(points):
+        # check increasing in fpr and tpr
+        last = None
+        for point in points:
+            logging.debug(point)
+            try:
+                assert point.fpr() <= last.fpr()
+                assert point.tpr() <= last.tpr()
+            except AttributeError:
+                pass
+            last = point
+
+        # check endpoints
+        assert 1 == points[ 0].fpr() == points[ 0].tpr()
+        assert 0 == points[-1].fpr() == points[-1].tpr()
 
     P.close('all')
-
-    roc_thresholds = create_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_points=32)
-    rocs = [roc for roc, t in roc_thresholds]
-    auc = area_under_curve(rocs)
-
     P.figure()
-    plot_roc_points(rocs, label='classifier : AUC=%.3f' % auc, color='blue', marker='s')
-    plot_random_classifier(label='Random')
-    label_plot()
+
+    # Try one positive, one negative, negative first
+    positive = [.5]
+    negative = [.5]
+    logging.info('Positive=%s; Negative=%s', positive, negative)
+    points = list(all_rocs_from_thresholds(positive, negative, True))
+    check_points(points)
+    logging.info(
+        'Got %4d ROC points, AUC=%.3f, AUC50=%.3f',
+        len(points), auc(points), auc50_from_rocpoints(points[::-1]))
+    assert 3 == len(points)
+
+    # Try one positive, one negative
+    positive = [.5]
+    negative = [.5]
+    logging.info('Positive=%s; Negative=%s', positive, negative)
+    points = list(all_rocs_from_thresholds(positive, negative, False))
+    check_points(points)
+    logging.info(
+        'Got %4d ROC points, AUC=%.3f, AUC50=%.3f',
+        len(points), auc(points), auc50_from_rocpoints(points[::-1]))
+#     for point in points:
+#         logging.info(point)
+    assert 2 == len(points)
+
+    # Try one positive, one negative
+    positive = [.5]
+    negative = [.6]
+    logging.info('Positive=%s; Negative=%s', positive, negative)
+    points = list(all_rocs_from_thresholds(positive, negative, False))
+    check_points(points)
+    logging.info(
+        'Got %4d ROC points, AUC=%.3f, AUC50=%.3f',
+        len(points), auc(points), auc50_from_rocpoints(points[::-1]))
+#     for point in points:
+#         logging.info(point)
+    assert 3 == len(points)
+
+    # Try a few
+    R.seed(1)
+    positive1 = R.normal(size=500, loc=.6, scale=.4)
+    positive1.sort()
+    positive2 = R.normal(size=500, loc=.8, scale=.8)
+    positive2.sort()
+    negative = R.normal(size=700, loc=.2, scale=.4)
+    negative.sort()
+    golden_ratio = (1 + math.sqrt(5)) / 2
+    P.figure(figsize=(6, 6))
+    plot_random_classifier()
+    colors = ['green', 'blue']
+    hatches = ['/', '\\']
+    for idx, positive in enumerate((positive1, positive2)):
+        color = colors[idx]
+        points = list(all_rocs_from_thresholds(positive, negative, True))
+        check_points(points)
+        method = idx + 1
+        label = 'method %d' % method
+        logging.info(
+            '%s: Got %4d ROC points, AUC=%.3f, AUC50=%.3f',
+            label, 
+            len(points), 
+            auc(points), 
+            auc50_from_rocpoints(points[::-1]))
+        plot_rocpoints(points, label=label, color=color)
+        restricted_points = list(restrict_false_positives(points[::-1]))
+        plot_rocpoints(
+            restricted_points, 
+            label=None,
+            fillargs={
+                'alpha' : 0.3,
+                #'hatch' : hatches[idx],
+                'edgecolor' : (0,0,0,0),
+                'facecolor' : color,
+                #'facecolor' : (0,0,0,0),
+            },
+            color=(0,0,0,0),
+        )
     P.legend(loc='lower right')
+    label_plot()
     P.savefig('output/ROC.eps')
     P.savefig('output/ROC.png')
-    P.show()
-    P.close()
+    P.savefig('output/ROC.pdf')
+    P.close('all')
 
-#       P.figure()
-#       plot_precision_versus_recall(rocs, label='classifier : AUC=%.3f' % auc, color='blue', marker='s')
-#       label_precision_versus_recall()
-#       P.legend(loc='lower left')
-#       P.savefig('output/Precision-versus-Recall.eps')
-#       P.savefig('output/Precision-versus-Recall.png')
-#       P.show()
 
-#       P.figure()
-#       plot_precision_recall(roc_thresholds)
-#       label_precision_recall()
-#       P.legend(loc='lower center')
-#       P.savefig('output/Precision-Recall.eps')
-#       P.savefig('output/Precision-Recall.png')
-#       P.show()
-
-    positive_thresholds, negative_thresholds = resize_negative_examples(positive_thresholds, negative_thresholds)
-    roc_thresholds = create_rocs_from_thresholds(positive_thresholds, negative_thresholds, num_points=32)
-    rocs = [roc for roc, t in roc_thresholds]
-    auc = area_under_curve(rocs)
-    P.figure()
-    plot_roc_points(rocs, label='classifier : AUC=%.3f' % auc, color='blue', marker='s')
-    plot_random_classifier(label='Random')
-    label_plot()
-    P.legend(loc='lower right')
-    P.savefig('output/ROC50.eps')
-    P.savefig('output/ROC50.png')
-    P.show()
-    P.close()
